from __future__ import annotations

import itertools
import math
from typing import Iterable, cast

import matplotlib.patches as patches
import numpy as np
from matplotlib.axes import Axes
from mpl_toolkits.mplot3d.art3d import Poly3DCollection


class _RegularPolygon:
    def __init__(self, *coordinates: float) -> None:
        self.coordinates = coordinates

    def __str__(self) -> str:
        return f"{self.coordinates}"

    def __repr__(self) -> str:
        return f"{self.coordinates}"

    def __add__(self, other: _RegularPolygon) -> _RegularPolygon:
        return self.__class__(
            *(i + j for i, j in zip(self.coordinates, other.coordinates))
        )

    def __sub__(self, other: _RegularPolygon) -> _RegularPolygon:
        return self.__class__(
            *(i - j for i, j in zip(self.coordinates, other.coordinates))
        )

    def __mul__(self, other: _RegularPolygon) -> _RegularPolygon:
        return self.__class__(
            *(i * j for i, j in zip(self.coordinates, other.coordinates))
        )

    def __pow__(self, other: _RegularPolygon) -> _RegularPolygon:
        return self.__class__(
            *(i**j for i, j in zip(self.coordinates, other.coordinates))
        )

    def __truediv__(self, other: _RegularPolygon) -> _RegularPolygon:
        return self.__class__(
            *(i / j for i, j in zip(self.coordinates, other.coordinates))
        )

    def __floordiv__(self, other: _RegularPolygon) -> _RegularPolygon:
        return self.__class__(
            *(i // j for i, j in zip(self.coordinates, other.coordinates))
        )

    def __mod__(self, other: _RegularPolygon) -> _RegularPolygon:
        return self.__class__(
            *(i % j for i, j in zip(self.coordinates, other.coordinates))
        )

    def __iter__(self) -> Iterable[float]:
        yield from self.coordinates


# Rod coordinates (1D)


class Rod(_RegularPolygon):
    def __init__(self, x: float) -> None:
        super().__init__(x)
        self.L = 1

    def neighbors(self) -> tuple[Rod, Rod]:
        x = self.coordinates[0]
        return Rod(x - 1), Rod(x + 1)


# TRIANGLE COORDINATES
class Triangle(_RegularPolygon):
    def __init__(self, col: float, row: float) -> None:
        super().__init__(col, row)
        self.L = 1  # Side length
        self.d = self.L / math.sqrt(3)  # Radius
        self.h = self.L / 2 * math.sqrt(3)  # Height
        self.a = self.h - self.d  # Distance from middle to base
        self.num_vertices = 3
        self.orientation = math.radians(180 * (col % 2 + row % 2))
        self.x_offset = 0.5 * self.L
        self.y_offset = self.a

    def plot(
        self,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float] = "C1",
        edgecolor: str | tuple[float, float, float, float] = (0, 0, 0, 1),
    ) -> None:
        ax.add_patch(
            patches.RegularPolygon(
                self.to_plot().coordinates,
                numVertices=self.num_vertices,
                radius=self.d,
                facecolor=facecolor,
                edgecolor=edgecolor,
                orientation=self.orientation,
            )
        )

    def to_plot(self) -> Square:
        c, r = self.coordinates
        x = c / 2
        # Only add a if either r or c is odd
        y = r * self.h + (self.a * (cast(int, r) % 2 ^ cast(int, c) % 2))
        return Square(x, y)

    def neighbors(self) -> tuple[Triangle, Triangle, Triangle]:
        c, r = self.coordinates
        if c % 2 == r % 2:
            return (Triangle(c + 1, r), Triangle(c, r - 1), Triangle(c - 1, r))
        return (Triangle(c, r + 1), Triangle(c + 1, r), Triangle(c - 1, r))

    def distance(self, other: Triangle) -> float:
        """Manhattan distance"""
        c1, r1 = self.coordinates
        c2, r2 = other.coordinates
        return (
            abs(r1 - r2)
            + abs(r1 + (-c1 - r1) / 2 - r2 - (-c2 - r2) / 2)
            + abs((c1 + r1) / 2 - (c2 + r2) / 2)
        )


# SQUARE COORDINATES
class Square(_RegularPolygon):
    def __init__(self, x: float, y: float) -> None:
        super().__init__(x, y)
        self.L = 1  # side length
        self.r = math.sin(math.pi / 4) * self.L  # Radius
        self.num_vertices = 4
        self.orientation = math.radians(45)
        self.x_offset = 0.5 * self.L
        self.y_offset = 0.5 * self.L

    def plot(
        self,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float] = "C1",
        edgecolor: str | tuple[float, float, float, float] = (0, 0, 0, 1),
    ) -> None:
        ax.add_patch(
            patches.RegularPolygon(
                self.to_plot().coordinates,
                numVertices=self.num_vertices,
                radius=self.r,
                facecolor=facecolor,
                edgecolor=edgecolor,
                orientation=self.orientation,
            )
        )

    def to_oddq(self) -> Oddq:
        x, y = self.coordinates
        c = 2 / 3 * x
        r = y / math.sqrt(3) - 0.5 * (c % 2)
        return Oddq(int(c), int(r))

    def to_triangle(self) -> Triangle:
        x, y = self.coordinates
        t = Triangle(0, 0)  # Dummy triangle to get height and a
        c = (x - 0.5) * 2
        r = y - t.a
        # Remainder will be 0 or 0.333.., so I can just round it down
        r = round(r / t.h)
        return Triangle(int(c), int(r))

    def to_plot(self) -> Square:
        x, y = self.coordinates
        return Square(x, y)

    def neighbors(
        self,
        directions: Iterable[int] = (0, 1, 2, 3),
    ) -> list[Square]:
        x, y = self.coordinates
        surroundings = ((-1, 0), (0, 1), (1, 0), (0, -1))
        result = []
        for direction in directions:
            i, j = surroundings[direction]
            result.append(Square(x + i, y + j))
        return result

    def distance(self, Square2):
        """Manhattan distance"""
        x1, y1 = self.coordinates
        x2, y2 = Square2.coordinates
        return abs(x1 - x2) + abs(y1 - y2)


# HEXAGON COORDINATES
class Oddq(_RegularPolygon):
    def __init__(self, col: float, row: float) -> None:
        super().__init__(col, row)
        self.L = 1
        self.radius = 1
        self.num_vertices = 6
        self.orientation = math.radians(30)

    def plot(
        self,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float] = "C1",
        edgecolor: str | tuple[float, float, float, float] = (0, 0, 0, 1),
    ) -> None:
        x, y = self.to_plot().coordinates
        ax.add_patch(
            patches.RegularPolygon(
                (x, y),
                numVertices=self.num_vertices,
                radius=self.radius,
                facecolor=facecolor,
                edgecolor=edgecolor,
                orientation=self.orientation,
            )
        )

    def to_plot(self) -> Square:
        c, r = self.coordinates
        x = 3 / 2 * c
        y = math.sqrt(3) * (r + 0.5 * (c % 2))  # If column is even, shift row
        return Square(float(x), float(y))

    def to_cube(self) -> Cube:
        c, r = self.coordinates
        x = c
        z = r - (c - c % 2) / 2
        y = -x - z
        return Cube(int(x), int(y), int(z))

    def to_axial(self) -> Axial:
        c, r = self.coordinates
        return Axial(int(c), int(r - (c - c % 2) / 2))

    def to_doubleheight(self) -> Doubleheight:
        c, r = self.coordinates
        return Doubleheight(int(c), int(r * 2 + (c % 2)))

    def neighbors(self, directions: Iterable[int] = (0, 1, 2, 3, 4, 5)) -> list[Oddq]:
        """Neighbors of a cell in odd-q coordinates.
        Coordinates start at the upper right and go clockwise
        """
        c, r = self.coordinates
        parity = cast(int, c % 2)
        surroundings = (
            ((1, 0), (1, -1), (0, -1), (-1, -1), (-1, 0), (0, 1)),
            ((1, 1), (1, 0), (0, -1), (-1, 0), (-1, 1), (0, 1)),
        )

        result = []
        for direction in directions:
            i, j = surroundings[parity][direction]
            result.append(Oddq(c + i, r + j))
        return result

    def distance(self, other: Oddq) -> float:
        """Manhattan distance"""
        return self.to_cube().distance(other.to_cube())


class Cube(_RegularPolygon):
    def __init__(self, x: float, y: float, z: float) -> None:
        super().__init__(x, y, z)

    def to_oddq(self) -> Oddq:
        x, _, z = self.coordinates
        c = x
        r = z + (x - x % 2) / 2
        return Oddq(int(c), int(r))

    def to_axial(self) -> Axial:
        x, _, z = self.coordinates
        return Axial(int(x), int(z))

    def to_doubleheight(self) -> Doubleheight:
        x, _, z = self.coordinates
        c = x
        r = 2 * z + x
        return Doubleheight(c, r)

    def neighbors(self, directions: Iterable[int] = (0, 1, 2, 3, 4, 5)) -> list[Cube]:
        """Neighbors of a cell in cube coordinates.
        Coordinates start at the upper right and go clockwise
        """
        x, y, z = self.coordinates
        surroundings = (
            (1, -1, 0),
            (1, 0, -1),
            (0, 1, -1),
            (-1, 1, 0),
            (-1, 0, 1),
            (0, -1, 1),
        )
        result = []
        for direction in directions:
            i, j, k = surroundings[direction]
            result.append(Cube(x + i, y + j, z + k))
        return result

    def distance(self, cube2: Cube) -> float:
        """Distance between two cells c1 and c2 in cube coordinates.
        In a cube grid, Manhattan distances are abs(dx) + abs(dy) + abs(dz).
        The distance on a hex grid is half that:
        d = (abs(x1 - x2) + abs(y1 - y1) + abs(z1 - z2)) / 2
        An equivalent way to write this is by noting that one of the three coordinates
        must be the sum of the other two, then picking that one as the distance:
        d = max(abs(x1 - x2), abs(y1 - y2), abs(z1 - z2))
        """
        x1, y1, z1 = self.coordinates
        x2, y2, z2 = cube2.coordinates
        return max(abs(x1 - x2), abs(y1 - y2), abs(z1 - z2))

    def ring(self, radius: int) -> list[Cube]:
        """Gives the coordinates of all cells surrounding the center cell with radius r
        Expects cube coordinates.
        """
        x, y, z = self.coordinates
        results = []
        # Go radius * steps in direction +z (4)
        x -= radius
        z += radius

        # Go around each edge
        for direction in range(6):
            # For radius steps
            for _ in range(radius):
                c = Cube(x, y, z)
                results.append(c)
                x, y, z = c.neighbors([direction])[0].coordinates
        return results


class Axial(_RegularPolygon):
    def __init__(self, q: float, r: float) -> None:
        super().__init__(q, r)

    def to_oddq(self) -> Oddq:
        q, r = self.coordinates
        return Oddq(int(q), int(r + (q - q % 2) / 2))

    def to_cube(self) -> Cube:
        q, r = self.coordinates
        return Cube(int(q), int(-q - r), int(r))

    def neighbors(self, directions: Iterable[int] = (0, 1, 2, 3, 4, 5)) -> list[Axial]:
        """Neighbors of a cell in axial coordinates
        Coordinates start at the upper right and go clockwise
        """
        q, r = self.coordinates
        surroundings = ((1, 0), (1, -1), (0, -1), (-1, 0), (-1, 1), (0, 1))
        result = []
        for direction in directions:
            i, j = surroundings[direction]
            result.append(Axial(q + i, r + j))
        return result

    def distance(self, other: Axial) -> float:
        """Manhattan distance"""
        return self.to_cube().distance(other.to_cube())


class Doubleheight(_RegularPolygon):
    def __init__(self, col: float, row: float) -> None:
        super().__init__(col, row)

    def to_oddq(self) -> Oddq:
        c, r = self.coordinates
        return Oddq(int(c), int((r - (c % 2)) / 2))

    def to_cube(self) -> Cube:
        c, r = self.coordinates
        x = c
        z = (r - c) / 2
        y = -x - z
        return Cube(x, y, z)

    def neighbors(
        self, directions: Iterable[int] = (0, 1, 2, 3, 4, 5)
    ) -> list[Doubleheight]:
        surroundings = (
            (+1, +1),
            (+1, -1),
            (0, -2),
            (-1, -1),
            (-1, +1),
            (0, +2),
        )
        c, r = self.coordinates
        result = []
        for direction in directions:
            i, j = surroundings[direction]
            result.append(Doubleheight(c + i, r + j))
        return result

    def distance(self, Doubleheight2):
        """Manhattan distance"""
        return self.to_cube().distance(Doubleheight2.to_cube())


class Cube3D(_RegularPolygon):
    def __init__(self, x: float, y: float, z: float) -> None:
        super().__init__(x, y, z)

    def to_plot(self) -> Cube3D:
        x, y, z = self.coordinates
        return Cube3D(x + 0.5, y + 0.5, z + 0.5)

    def plot(
        self,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float] = "C1",
        edgecolor: str | tuple[float, float, float, float] = (0, 0, 0, 1),
    ) -> None:
        cube = [
            [[0, 1, 0], [0, 0, 0], [1, 0, 0], [1, 1, 0]],
            [[0, 0, 0], [0, 0, 1], [1, 0, 1], [1, 0, 0]],
            [[1, 0, 1], [1, 0, 0], [1, 1, 0], [1, 1, 1]],
            [[0, 0, 1], [0, 0, 0], [0, 1, 0], [0, 1, 1]],
            [[0, 1, 0], [0, 1, 1], [1, 1, 1], [1, 1, 0]],
            [[0, 1, 1], [0, 0, 1], [1, 0, 1], [1, 1, 1]],
        ]

        cube = np.array(cube) + np.array(self.coordinates)

        ax.add_collection3d(  # type: ignore
            Poly3DCollection(
                cube,
                facecolors=list(itertools.repeat(facecolor, 6)),
                edgecolor=edgecolor,
            )
        )

    def neighbors(self, directions: Iterable[int] = (0, 1, 2, 3, 4, 5)) -> list[Cube3D]:
        x, y, z = self.coordinates
        surroundings = (
            (-1, 0, 0),
            (0, 1, 0),
            (1, 0, 0),
            (0, -1, 0),
            (0, 0, 1),
            (0, 0, -1),
        )
        result = []
        for direction in directions:
            i, j, k = surroundings[direction]
            result.append(Cube3D(x + i, y + j, z + k))
        return result



from __future__ import annotations

from typing import Any

from pybind11.setup_helpers import Pybind11Extension, build_ext


def build(setup_kwargs: dict[Any, Any]) -> None:
    ext_modules = [
        Pybind11Extension(
            "dismo.utils._perffuncs",
            ["src/dismo/utils/perffuncs.cpp"],
        ),
    ]
    setup_kwargs.update(
        {
            "ext_modules": ext_modules,
            "cmd_class": {"build_ext": build_ext},
            "zip_safe": False,
        }
    )



from __future__ import annotations

import copy
import itertools
import math
from typing import Any, Type
import warnings

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from numpy.typing import NDArray

try:
    from assimulo.problem import Explicit_Problem
    from assimulo.solvers import CVode

    ASSIMULO_SUPPORT_FLAG = True

    def _simulate(
        self: GridModel, y0: NDArray, t_end: float, t_points: int = 0
    ) -> tuple[NDArray, NDArray]:
        if t_points > 0:
            t_points -= 1
        problem = Explicit_Problem(self.get_right_hand_side, y0)
        integrator = CVode(problem)
        integrator.verbosity = 50
        t, y = integrator.simulate(t_end, t_points)
        return np.array(t), np.array(y)

except ImportError:
    import scipy.integrate

    def _simulate(
        self: GridModel, y0: NDArray, t_end: float, t_points: int = 0
    ) -> tuple[NDArray, NDArray]:
        if t_points > 0:
            t = np.linspace(0, t_end, t_points)
        else:
            t = np.linspace(0, t_end, 100)
        y = scipy.integrate.odeint(self.get_right_hand_side, y0, t, tfirst=True)
        return np.array(t), np.array(y)

    ASSIMULO_SUPPORT_FLAG = False

from .utils.coordinates import Cube3D, Oddq, Rod, Square, Triangle


class GridModel:
    """General grid model representation"""

    def __init__(
        self,
        gridsize: tuple[int, ...],
        cells: NDArray | None = None,
        parameters: dict[str, float] | None = None,
    ) -> None:
        warnings.warn("Experimental class, API might change.", FutureWarning)
        self._base_shape = None
        self._n_neighbors = 0
        self.initialized = True
        self.gridsize = gridsize
        if cells is None:
            self.cells = np.zeros(gridsize, dtype="int")
        else:
            self.cells = cells
        self.celltypes: dict[str, int] = {}
        self.celltype_indices: dict[str, list[int]] = {}
        self._is_celltype_array: dict[str, list[bool]] = {}
        self.neighbors: dict[str, int] = {}
        if parameters is None:
            self.parameters = {}
        else:
            self.parameters = parameters
        self.variables: list[str] = []

        if not ASSIMULO_SUPPORT_FLAG:
            warnings.warn("Assimulo not found, disabling sundials support.")

    def __reduce__(
        self,
    ) -> tuple[Type[GridModel], tuple[tuple[int, ...], dict[str, float]]]:
        return (self.__class__, (self.gridsize, self.parameters))

    def copy(self) -> GridModel:
        return copy.deepcopy(self)

    def add_celltype(self, name: str, value: int) -> None:
        self.celltypes[name] = value
        self.celltype_indices[name] = []
        self._is_celltype_array[name] = []
        self.initialized = False

    def add_cell(self, coordinates: tuple[int, int], celltype: str) -> None:
        self.cells[coordinates] = self.celltypes[celltype]
        self.initialized = False

    def remove_cell(self, coordinates: tuple[int, int]) -> None:
        self.cells[coordinates] = 0
        self.initialized = False

    def add_variable(self, variable_name: str) -> None:
        if variable_name not in self.variables:
            self.variables.append(variable_name)
            self.initialized = False

    def get_celltypes(self) -> dict[str, int]:
        return self.celltypes

    def get_index_of_coordinates(self, coordinates: tuple[int, int]) -> Any:
        return np.ravel_multi_index(coordinates, self.gridsize)

    def get_coordinates_of_index(self, index: int) -> Any:
        return np.unravel_index(index, self.gridsize)

    def get_indices_of_celltype(self, celltype: str) -> list[int]:
        type_value = self.celltypes[celltype]
        return [
            self.get_index_of_coordinates(i)  # type: ignore
            for i in zip(*np.where(self.cells == type_value))
        ]

    def get_celltype_of_index(self, index: int) -> NDArray:
        return self.cells[self.get_coordinates_of_index(index)]

    def get_cell_neighbors(self, idx):
        """Fills empty positions with -1"""
        upper_borders = self.cells.shape
        neighbors = np.full(self._n_neighbors, -1)
        for idx, neighbor in enumerate(
            self._base_shape(*self.get_coordinates_of_index(idx)).neighbors()
        ):
            neighbor_coordinates = neighbor.coordinates
            if any((i < 0 for i in neighbor_coordinates)):
                pass
            elif any(
                (i >= j for i, j in zip(neighbor_coordinates, upper_borders))
            ):
                pass
            elif self.cells[neighbor_coordinates] != 0:
                neighbors[idx] = self.get_index_of_coordinates(
                    neighbor_coordinates
                )
        return neighbors

    def get_cell_neighbors_of_celltype(self, idx, celltype):
        celltype = self.celltypes[celltype]
        neighbors = np.full(self._n_neighbors, -1)
        for idx, i in enumerate(self.get_cell_neighbors(idx)):
            if i != -1:
                if self.cells[self.get_coordinates_of_index(i)] == celltype:
                    neighbors[idx] = i
        return neighbors

    def get_all_neighbors_of_celltype(self, celltype, neighbor_celltype):
        arr = np.full((np.product(self.gridsize), self._n_neighbors), -1)
        for idx in self.get_indices_of_celltype(celltype):
            arr[idx] = self.get_cell_neighbors_of_celltype(
                idx, neighbor_celltype
            )
        return arr

    def nearest_neighbor_of_celltype(self, cell, neighbor_celltype):
        start = self._base_shape(*cell)
        distances = []
        for c, r in zip(
            *np.where(self.cells == self.celltypes[neighbor_celltype])
        ):
            distances.append(start.distance(self._base_shape(c, r)))
        return min(distances)

    def nearest_distances_of_celltypes(self, celltype1, celltype2):
        distances = []
        for c, r in zip(*np.where(self.cells == self.celltypes[celltype1])):
            distances.append(
                self.nearest_neighbor_of_celltype((c, r), celltype2)
            )
        return distances

    def initialize(self):
        """
        All functions need boolean arrays of cells
        Diffusion and active transport further need neighbor arrays
        """
        # Generate celltype indices
        for celltype in self.celltypes:
            self.celltype_indices[celltype] = np.array(
                self.get_indices_of_celltype(celltype)
            )
            self._is_celltype_array[celltype] = (
                self.cells.flatten() == self.celltypes[celltype]
            )

        # Generate neighbor arrays
        for in_celltype in self.celltypes:
            for out_celltype in self.celltypes:
                self.neighbors[
                    in_celltype, out_celltype
                ] = self.get_all_neighbors_of_celltype(
                    in_celltype, out_celltype
                )
        self.initialized = True

    def generate_initial_values(self):
        # Let's start with a simple: Everything is zero ;)
        y0 = {
            var: np.zeros(self.gridsize, dtype="float")
            for var in self.variables
        }
        return np.stack(tuple(y0.values())).flatten()

    def get_right_hand_side(self, t, y0):
        raise NotImplementedError

    def simulate(self, y0, t_end, t_points=0):
        t, y = _simulate(self, y0, t_end, t_points)
        n_vars = len(self.variables)
        n_dims = self.cells.shape
        if n_vars < 2:
            return t, y.reshape(len(t), *n_dims)
        else:
            y = y.reshape(len(t), n_vars, *n_dims)
            return (t, *[np.squeeze(i) for i in np.split(y, n_vars, axis=1)])

    def plot_axes(self):
        raise NotImplementedError

    def plot_cell(
        self,
        c,
        r,
        ax,
        facecolor: str | tuple[float, float, float, float] = "C1",
        edgecolor: str | tuple[float, float, float, float] = (0, 0, 0, 1),
    ):
        self._base_shape(c, r).plot(
            ax=ax, facecolor=facecolor, edgecolor=edgecolor
        )

    def plot_grid(self, ax):
        cols, rows = self.cells.shape
        for c, r in itertools.product(range(cols), range(rows)):
            self.plot_cell(c, r, facecolor=(0.50, 0.50, 0.50, 1 / 16), ax=ax)

    def plot_cell_coordinates(self, ax, fontsize=14):
        for coordinates in itertools.product(
            *(range(var) for var in self.cells.shape)
        ):
            ax.text(
                *self._base_shape(*coordinates).to_plot(),
                f"{coordinates}",
                fontsize=fontsize,
                ha="center",
                va="center",
            )

    def plot_cell_concentration(self, ax, concentrations, **kwargs):
        local_kwargs = {"ha": "center", "fontsize": 12}
        if kwargs:
            local_kwargs.update(kwargs)
        for coordinates in zip(*np.where(self.cells != 0)):
            ax.text(
                *self._base_shape(*coordinates).to_plot(),
                f"{concentrations[coordinates]:.2g}",
                **local_kwargs,
            )


class RodGridModel(GridModel):
    def __init__(self, gridsize, cells=None, parameters=None) -> None:
        super().__init__(gridsize, cells=cells, parameters=parameters)
        self._base_shape = Rod
        self._n_neighbors = 2


class TriGridModel(GridModel):
    def __init__(self, gridsize, cells=None, parameters=None) -> None:
        super().__init__(gridsize, cells=cells, parameters=parameters)
        self._base_shape = Triangle
        self._n_neighbors = 3

    def plot_axes(self, figsize=None, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize)
        else:
            fig = ax.get_figure()
        ax.set_aspect("equal")

        c, r = self.cells.shape
        base_shape = self._base_shape(0, 0)
        x_offset = base_shape.x_offset
        y_offset = base_shape.y_offset
        x_lower = -x_offset
        x_upper = 0.5 * c
        y_lower = -y_offset
        y_upper = r * base_shape.h - y_offset
        ax.set_xlim(x_lower, x_upper)
        ax.set_ylim(y_lower, y_upper)
        ax.axis("off")
        return fig, ax


class SquareGridModel(GridModel):
    def __init__(self, gridsize, cells=None, parameters=None) -> None:
        super().__init__(gridsize, cells=cells, parameters=parameters)
        self._base_shape = Square
        self._n_neighbors = 4

    def plot_axes(self, figsize=None, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize)
        else:
            fig = ax.get_figure()
        ax.set_aspect("equal")

        c, r = self.cells.shape
        x_offset = self._base_shape(0, 0).x_offset
        y_offset = self._base_shape(0, 0).y_offset

        x_lower = -x_offset - 0.1
        x_upper = c - x_offset + 0.1
        y_lower = -y_offset - 0.1
        y_upper = r - y_offset + 0.1
        ax.set_xlim(x_lower, x_upper)
        ax.set_ylim(y_lower, y_upper)
        ax.axis("off")
        return fig, ax


class HexGridModel(GridModel):
    def __init__(self, gridsize, cells=None, parameters=None) -> None:
        super().__init__(gridsize, cells=cells, parameters=parameters)
        self._base_shape = Oddq
        self._n_neighbors = 6

    def plot_axes(self, figsize=None, ax=None):
        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=figsize)
        else:
            fig = ax.get_figure()
        ax.set_aspect("equal")

        c, r = self.cells.shape
        x_lower = -1
        x_upper = 1 + 3 / 2 * (c - 1)
        y_lower = -math.sqrt(3) * 0.5
        y_upper = math.sqrt(3) * r
        ax.set_xlim(x_lower, x_upper)
        ax.set_ylim(y_lower, y_upper)
        ax.axis("off")
        return fig, ax


class CubeGridModel(GridModel):
    def __init__(self, gridsize, cells=None, parameters=None) -> None:
        super().__init__(gridsize, cells=cells, parameters=parameters)
        self._base_shape = Cube3D
        self._n_neighbors = 6

    def plot_axes(self, figsize=None, ax=None):
        if ax is None:
            fig, ax = plt.subplots(
                1, 1, figsize=figsize, subplot_kw={"projection": "3d"}
            )
        else:
            fig = ax.get_figure()

        a, b, c = self.cells.shape
        ax.set_xlim(0, a)
        ax.set_ylim(0, b)
        ax.set_zlim(0, c)
        ax.axis("off")
        return fig, ax

    def plot_cell(
        self,
        c: float,
        r: float,
        z: float,
        ax: Axes,
        facecolor: str | tuple[float, float, float, float] = "C1",
        edgecolor: str | tuple[float, float, float, float] = (0, 0, 0, 1 / 8),
    ):
        self._base_shape(c, r, z).plot(
            ax=ax, facecolor=facecolor, edgecolor=edgecolor
        )
        return ax

    def plot_grid(
        self,
        ax,
        facecolor: str
        | tuple[float, float, float, float] = (0.50, 0.50, 0.50, 1 / 16),
        edgecolor: str | tuple[float, float, float, float] = (0, 0, 0, 1 / 16),
    ):
        x, y, z = self.cells.shape
        for x, y, z in itertools.product(range(x), range(y), range(z)):
            self.plot_cell(
                x, y, z, facecolor=facecolor, edgecolor=edgecolor, ax=ax
            )
        return ax



from __future__ import annotations

import warnings

try:
    from ._perffuncs import (  # type: ignore
        active_transport,
        diffusion,
        diffusion_nonavg,
    )
except ImportError:
    warnings.warn("Import failed, falling back to Python diffusion function")

    def diffusion(
        dydt: list[float],
        y0: list[float],
        is_celltype: list[bool],
        celltype_neighbors: list[list[int]],
        alpha: float,
        n_neighbors: float,
    ):
        for cell, cell_conc in enumerate(y0):
            if is_celltype[cell]:
                diff = 0.0
                for neighbor in celltype_neighbors[cell]:
                    if neighbor != -1:
                        diff += y0[neighbor] - cell_conc
                dydt[cell] += diff / n_neighbors * alpha
        return dydt

    def diffusion_nonavg(
        dydt: list[float],
        y0: list[float],
        is_celltype: list[bool],
        celltype_neighbors: list[list[int]],
        alpha: float,
        n_neighbors: float,
    ):
        for cell, cell_conc in enumerate(y0):
            if is_celltype[cell]:
                diff = 0.0
                for neighbor in celltype_neighbors[cell]:
                    if neighbor != -1:
                        diff += y0[neighbor] - cell_conc
                dydt[cell] += diff * alpha
        return dydt

    def active_transport(
        dydt: list[float],
        y0: list[float],
        is_celltype: list[bool],
        celltype_neighbors: list[list[int]],
        alpha: float,
        n_neighbors: float,
    ):
        for cell, cell_conc in enumerate(y0):
            if is_celltype[cell]:
                for neighbor in celltype_neighbors[cell]:
                    if neighbor != -1:
                        diff = (y0[neighbor] - cell_conc) * alpha
                        if diff < 0:
                            dydt[neighbor] += diff
                            dydt[cell] -= diff
        return dydt



from __future__ import annotations

import math
from typing import Any

import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import animation
from matplotlib.animation import FuncAnimation
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from numpy.typing import NDArray

from .coordinates import Oddq


def plot_leaf_axes(
    leaf: Any,
    figsize: tuple[float, float],
    ax: Axes | None = None,
) -> tuple[Figure, Axes]:
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=figsize)
    else:
        fig = ax.get_figure()
    ax.set_aspect("equal")

    c, r = leaf.cells.shape
    x_lower = -1
    x_upper = 1 + 3 / 2 * (c - 1)
    y_lower = -math.sqrt(3) * 0.5
    y_upper = math.sqrt(3) * r
    ax.set_xlim(x_lower, x_upper)
    ax.set_ylim(y_lower, y_upper)
    ax.set_xticks([])
    ax.set_yticks([])
    return fig, ax


def plot_leaf_cells(
    leaf: Any,
    ax: Axes,
    concentrations: NDArray | None,
    min_conc: float | None = None,
    max_conc: float | None = None,
    alpha: float | None = None,
) -> Axes:
    if concentrations is None:
        concentrations = np.ones(leaf.gridsize)
        min_conc = 0
        max_conc = 1
    else:
        if min_conc is None:
            min_conc = np.min(
                concentrations[leaf.cells != 0]
            )  # Otherwise this is always 0
        if max_conc is None:
            max_conc = np.max(concentrations)
        assert min_conc is not None
        assert max_conc is not None
        if min_conc == max_conc:
            max_conc += 0.1

    normalized_concentrations = (concentrations - min_conc) / (
        max_conc - min_conc
    )

    def add_hexagon(
        c: float, r: float, color: str, alpha: float, ax: Axes = ax
    ) -> None:
        ax.add_patch(
            patches.RegularPolygon(
                Oddq(c, r).to_plot().coordinates,
                numVertices=6,
                radius=1,
                facecolor=color,
                alpha=alpha,
                orientation=np.radians(30),
                edgecolor="k",
            )
        )

    # Photosynthetic cells
    for c, r in zip(*np.where(leaf.cells == 1)):  # type: ignore
        if alpha is None:
            al = max(normalized_concentrations[c, r], 0.05)
        else:
            al = alpha
        add_hexagon(c, r, "green", al)

    # Veins
    for c, r in zip(*np.where(leaf.cells == 2)):  # type: ignore
        if alpha is None:
            al = max(normalized_concentrations[c, r], 0.05)
        else:
            al = alpha
        add_hexagon(c, r, "brown", al)

    # Stoma
    for c, r in zip(*np.where(leaf.cells == 3)):  # type: ignore
        if alpha is None:
            al = max(normalized_concentrations[c, r], 0.05)
        else:
            al = alpha
        add_hexagon(c, r, "blue", al)
    return ax


def plot_leaf_cell_annotations(
    leaf: Any,
    ax: Axes,
    concentrations: NDArray,
    **kwargs: dict[Any, Any],
):
    local_kwargs = {"ha": "center", "fontsize": 12}
    if kwargs:
        local_kwargs.update(kwargs)
    for c, r in zip(*np.where(leaf.cells != 0)):  # type: ignore
        ax.annotate(
            f"{concentrations[c, r]:.2g}",
            Oddq(c, r).to_plot().coordinates,
            **local_kwargs,
        )


def plot_leaf_cell_coordinates(
    leaf: Any,
    ax: Axes,
) -> None:
    for c, r in zip(*np.where(leaf.cells != 0)):  # type: ignore
        ax.annotate(f"{c, r}", Oddq(c, r).to_plot().coordinates, ha="center")


def plot_leaf(
    leaf: Any,
    concentrations: NDArray | None = None,
    min_conc: float | None = None,
    max_conc: float | None = None,
    figsize: tuple[float, float] = (10, 10),
    ax: Axes | None = None,
    annotate: bool = False,
    alpha: float | None = None,
    **kwargs,
):
    fig, ax = plot_leaf_axes(leaf, figsize=figsize, ax=ax)
    ax = plot_leaf_cells(
        leaf, ax, concentrations, min_conc, max_conc, alpha, **kwargs
    )
    if annotate:
        assert concentrations is not None
        ax = plot_leaf_cell_annotations(leaf, ax, concentrations, **kwargs)
    return fig, ax


def time_lapse(
    leaf: Any,
    y: list[NDArray],
    figsize: tuple[float, float] = (10, 10),
    ffmpeg_path: str = "/usr/bin/ffmpeg",
) -> FuncAnimation:
    plt.rcParams["animation.ffmpeg_path"] = ffmpeg_path
    fig, ax = plot_leaf_axes(leaf=leaf, figsize=figsize)
    plt.close()

    def init():
        ax.patches = []  # type: ignore
        return fig, ax

    def update_func(frame):
        ax.patches = []  # type: ignore
        ax.texts = []  # type: ignore
        y_current = y[frame]
        plot_leaf_cells(leaf=leaf, ax=ax, concentrations=y_current)
        plot_leaf_cell_annotations(leaf=leaf, ax=ax, concentrations=y_current)
        return None

    anim = animation.FuncAnimation(
        fig,
        update_func,
        frames=list(range(len(y) // 2)),
        init_func=init,
        repeat=False,
    )
    return anim



from __future__ import annotations

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import animation

from .basemodels import (
    CubeGridModel,
    HexGridModel,
    SquareGridModel,
    TriGridModel,
)
from .utils import perffuncs
from matplotlib.axes import Axes


def photosynthesis_constant(_: float, k: float) -> float:
    return k


def photosynthesis_saturating(sugar: float, km: float, vmax: float) -> float:
    return vmax / np.exp(km * sugar)


def photosynthesis_co2(sugar: float, co2: float, vmax: float) -> float:
    return vmax * co2 / np.exp(sugar)


def photosynthesis_reversible(
    sugar: float, co2: float, midpoint: float, magnitude: float
) -> float:
    return (magnitude / (1 + np.exp(sugar - midpoint)) - magnitude / 2) * co2


def co2_influx(co2: float, x0: float = 10, L: float = 1) -> float:
    """Self-limiting co2 influx function"""
    return L / (1 + np.exp(co2 - x0)) - L / 2


def outflux(sugar: float, alpha: float) -> float:
    return -sugar * alpha


class _Leaf:
    def __init__(self) -> None:
        self.influx_functions = {}
        self.single_cell_outflux_functions = {}
        self.outflux_functions = {}
        self.diffusion_functions = {}
        self.active_transport = {}

    def add_influx_function(
        self,
        name,
        in_vars,
        out_vars,
        input_cells,
        output_cells,
        function,
        parameters,
    ):
        self.influx_functions[name] = {
            "func": function,
            "in_vars": in_vars,
            "out_vars": out_vars,
            "pars": parameters,
            "in_cells": input_cells,
            "out_cells": output_cells,
        }

    def add_outflux_function(
        self, name, out_vars, output_cells, function, parameters
    ):
        self.outflux_functions[name] = {
            "func": function,
            "out_vars": out_vars,
            "pars": parameters,
            "out_cells": output_cells,
        }

    def add_single_cell_outflux_function(
        self, name, cell_index, var_type, function, parameters
    ):
        self.single_cell_outflux_functions[name] = {
            "func": function,
            "cell_index": cell_index,
            "var_type": var_type,
            "pars": parameters,
        }

    def add_diffusion_function(
        self, name, variable, in_cells, out_cells, parameter
    ):
        self.diffusion_functions[name] = {
            "variable": variable,
            "in_cells": in_cells,
            "out_cells": out_cells,
            "parameter": parameter,
        }

    def add_active_transport_function(
        self, name, variable, in_cells, out_cells, parameter
    ):
        self.active_transport[name] = {
            "variable": variable,
            "in_cells": in_cells,
            "out_cells": out_cells,
            "parameter": parameter,
        }

    def get_right_hand_side(self, t, y):
        # y
        in_vars = dict(zip(self.variables, np.hsplit(y, len(self.variables))))
        # dydt
        out_vars = {
            i: np.zeros(self.gridsize).flatten() for i in self.variables
        }

        # Influx functions
        for influx_function in self.influx_functions.values():
            (
                func,
                in_types,
                out_type,
                pars,
                in_cells,
                out_cells,
            ) = influx_function.values()
            out_vars[out_type][self.celltype_indices[out_cells]] += func(
                *[
                    in_vars[type_][self.celltype_indices[in_cells]]
                    for type_ in in_types
                ],
                *[self.parameters[i] for i in pars],
            )

        # Diffusion functions
        for func in self.diffusion_functions.values():
            variable, in_cells, out_cells, parameter = func.values()
            out_vars[variable] = self.diffusion_function(
                out_vars[variable],
                in_vars[variable],
                self._is_celltype_array[in_cells],
                self.neighbors[in_cells, out_cells],
                self.parameters[parameter],
                self._n_neighbors,
            )

        # Active transport functions
        for func in self.active_transport.values():
            variable, in_cells, out_cells, parameter = func.values()
            out_vars[variable] = self.active_transport_function(
                out_vars[variable],
                in_vars[variable],
                self._is_celltype_array[in_cells],
                self.neighbors[in_cells, out_cells],
                self.parameters[parameter],
                self._n_neighbors,
            )

        # Single cell outflux functions
        for outflux_func in self.single_cell_outflux_functions.values():
            func, cell_index, var_type, pars = outflux_func.values()
            out_vars[var_type][cell_index] += func(
                in_vars[var_type][cell_index],
                *[self.parameters[i] for i in pars],
            )

        # Outflux functions
        for outflux_function in self.outflux_functions.values():
            func, in_types, pars, in_cells = outflux_function.values()
            out_vars[in_types][self.celltype_indices[in_cells]] += func(
                *[
                    in_vars[type_][self.celltype_indices[in_cells]]
                    for type_ in in_types
                ],
                *[self.parameters[i] for i in pars],
            )
        return np.array(list(out_vars.values())).flatten()

    def plot_cells(
        self,
        ax: Axes,
        concentrations=None,
        min_conc=None,
        max_conc=None,
        alpha=None,
        **kwargs,
    ):
        if concentrations is None:
            concentrations = np.ones(self.gridsize)
            min_conc = 0
            max_conc = 1
        else:
            if min_conc is None:
                min_conc = np.min(
                    concentrations[self.cells != 0]
                )  # Otherwise this is always 0
            if max_conc is None:
                max_conc = np.max(concentrations)
            if min_conc == max_conc:
                max_conc += 0.1
        normalized_concentrations = (concentrations - min_conc) / (
            max_conc - min_conc
        )

        # Photosynthetic cells
        for c, r in zip(*np.where(self.cells == 1)):
            if alpha is None:
                al = max(normalized_concentrations[c, r], 0.05)
            else:
                al = alpha
            self.plot_cell(
                c,
                r,
                facecolor=(0.00, 0.50, 0.00, al),
                edgecolor=(0, 0, 0, 1),
                ax=ax,
            )

        # Veins
        for c, r in zip(*np.where(self.cells == 2)):
            if alpha is None:
                al = max(normalized_concentrations[c, r], 0.05)
            else:
                al = alpha
            self.plot_cell(
                c,
                r,
                facecolor=(0.30, 0.15, 0.03, al),
                edgecolor=(0, 0, 0, 1),
                ax=ax,
            )

        # Stoma
        for c, r in zip(*np.where(self.cells == 3)):
            if alpha is None:
                al = max(normalized_concentrations[c, r], 0.05)
            else:
                al = alpha
            self.plot_cell(
                c,
                r,
                facecolor=(0.80, 0.37, 0.04, al),
                edgecolor=(0, 0, 0, 1),
                ax=ax,
            )
        return ax

    def plot(
        self,
        concentrations=None,
        min_conc=None,
        max_conc=None,
        figsize=(10, 10),
        ax: Axes | None = None,
        annotate=False,
        alpha=None,
        **kwargs,
    ):
        fig, ax = self.plot_axes(figsize=figsize, ax=ax)
        ax = self.plot_cells(
            ax, concentrations, min_conc, max_conc, alpha, **kwargs
        )
        if annotate:
            ax = self.plot_cell_concentrations(ax, concentrations, **kwargs)
        return fig, ax

    def time_lapse(
        self,
        y,
        annotate=False,
        filename=None,
        figsize=(10, 10),
        ffmpeg_path="/usr/bin/ffmpeg",
    ):
        plt.rcParams["animation.ffmpeg_path"] = ffmpeg_path
        fig, ax = self.plot_axes(figsize=figsize)
        plt.close()

        def init():
            ax.patches = []
            return fig, ax

        def update_func(frame):
            ax.patches = []
            ax.texts = []
            y_current = y[frame]
            self.plot_cells(ax=ax, concentrations=y_current)
            if annotate:
                self.plot_cell_concentration(ax=ax, concentrations=y_current)
            return None

        anim = animation.FuncAnimation(
            fig,
            update_func,
            frames=list(range(len(y) // 2)),
            init_func=init,
            repeat=False,
        )
        return anim


class _VeinModel:
    def __init__(self, vein_base_coordinates) -> None:
        self.add_celltype("mesophyll", 1)
        self.add_celltype("vein", 2)
        self.add_variable("sucrose")
        self._vein_base = self.get_index_of_coordinates(vein_base_coordinates)

        self.add_influx_function(
            "ps",
            in_vars=["sucrose"],
            out_vars="sucrose",
            input_cells="mesophyll",
            output_cells="mesophyll",
            function=photosynthesis_saturating,
            parameters=["ps_km", "ps_vmax"],
        )
        self.add_diffusion_function(
            name="suc_meso_to_meso",
            variable="sucrose",
            in_cells="mesophyll",
            out_cells="mesophyll",
            parameter="suc_meso_to_meso",
        )
        self.add_diffusion_function(
            name="suc_vein_to_vein",
            variable="sucrose",
            in_cells="vein",
            out_cells="vein",
            parameter="suc_vein_to_vein",
        )
        self.add_active_transport_function(
            name="suc_meso_to_vein",
            variable="sucrose",
            in_cells="mesophyll",
            out_cells="vein",
            parameter="suc_meso_to_vein",
        )
        self.add_single_cell_outflux_function(
            name="main_vein",
            cell_index=self._vein_base,
            var_type="sucrose",
            function=outflux,
            parameters=["suc_vein_to_vein"],
        )

    def get_vein_outflux(self, y):
        """Get vein outflux for a single concentration array"""
        func, cell_idx, var_type, pars = self.single_cell_outflux_functions[
            "main_vein"
        ].values()
        if y.ndim == 2:
            return (
                func(y.flatten()[cell_idx], *[self.parameters[i] for i in pars])
                * -1
            )
        else:
            res = [
                func(i.flatten()[cell_idx], *[self.parameters[i] for i in pars])
                for i in y
            ]
            return np.array(res) * -1


class _StomataModel(_VeinModel):
    def __init__(self, vein_base_coordinates) -> None:
        super().__init__(vein_base_coordinates=vein_base_coordinates)
        self.add_celltype("stoma", 3)
        self.add_variable("co2")

        self.add_influx_function(
            "ps",
            in_vars=["sucrose", "co2"],
            out_vars="sucrose",
            input_cells="mesophyll",
            output_cells="mesophyll",
            function=photosynthesis_co2,
            parameters=["ps_vmax"],
        )
        self.add_influx_function(
            "co2_influx",
            in_vars=["co2"],
            out_vars="co2",
            input_cells="stoma",
            output_cells="stoma",
            function=photosynthesis_saturating,
            parameters=["co2_km", "co2_vmax"],
        )
        self.add_active_transport_function(
            name="co2_stoma_to_meso",
            variable="co2",
            in_cells="stoma",
            out_cells="mesophyll",
            parameter="co2_stoma_to_meso",
        )
        self.add_diffusion_function(
            name="co2_meso_to_meso",
            variable="co2",
            in_cells="mesophyll",
            out_cells="mesophyll",
            parameter="co2_meso_to_meso",
        )


class TriLeafModel(_Leaf, TriGridModel):
    def __init__(self, gridsize, cells=None, parameters=None) -> None:
        TriGridModel.__init__(
            self, gridsize, cells=cells, parameters=parameters
        )
        _Leaf.__init__(self)
        self.diffusion_function = perffuncs.diffusion_nonavg
        self.active_transport_function = perffuncs.active_transport


class SquareLeafModel(_Leaf, SquareGridModel):
    def __init__(self, gridsize, cells=None, parameters=None) -> None:
        SquareGridModel.__init__(
            self, gridsize, cells=cells, parameters=parameters
        )
        _Leaf.__init__(self)
        self.diffusion_function = perffuncs.diffusion_nonavg
        self.active_transport_function = perffuncs.active_transport


class HexLeafModel(_Leaf, HexGridModel):
    def __init__(self, gridsize, cells=None, parameters=None) -> None:
        HexGridModel.__init__(
            self, gridsize, cells=cells, parameters=parameters
        )
        _Leaf.__init__(self)
        self.diffusion_function = perffuncs.diffusion_nonavg
        self.active_transport_function = perffuncs.active_transport


class CubeLeafModel(_Leaf, CubeGridModel):
    def __init__(self, gridsize, cells=None, parameters=None) -> None:
        CubeGridModel.__init__(
            self, gridsize, cells=cells, parameters=parameters
        )
        _Leaf.__init__(self)
        self.diffusion_function = perffuncs.diffusion_nonavg

    def plot_cells(
        self,
        ax: Axes,
        concentrations=None,
        min_conc=None,
        max_conc=None,
        alpha=None,
        **kwargs,
    ):
        # Normalize concentrations
        if concentrations is None:
            concentrations = np.ones(self.gridsize)
            min_conc = 0
            max_conc = 1
        else:
            if min_conc is None:
                min_conc = np.min(
                    concentrations[self.cells != 0]
                )  # Otherwise this is always 0
            if max_conc is None:
                max_conc = np.max(concentrations)
            if min_conc == max_conc:
                max_conc += 0.1
        normalized_concentrations = (concentrations - min_conc) / (
            max_conc - min_conc
        )

        for x, y, z in zip(*np.where(self.cells == 1)):
            if alpha is None:
                al = max(normalized_concentrations[x, y, z], 0)
            else:
                al = alpha
            self.plot_cell(
                x,
                y,
                z,
                facecolor=(0.00, 0.50, 0.00, al),
                edgecolor=(0, 0, 0, 1),
                ax=ax,
            )

        for x, y, z in zip(*np.where(self.cells == 2)):
            if alpha is None:
                al = max(normalized_concentrations[x, y, z], 0)
            else:
                al = alpha
            self.plot_cell(
                x,
                y,
                z,
                facecolor=(0.30, 0.15, 0.03, al),
                edgecolor=(0, 0, 0, 1),
                ax=ax,
            )

        for x, y, z in zip(*np.where(self.cells == 3)):
            if alpha is None:
                al = max(normalized_concentrations[x, y, z], 0)
            else:
                al = alpha
            self.plot_cell(
                x,
                y,
                z,
                facecolor=(0.80, 0.37, 0.04, al),
                edgecolor=(0, 0, 0, 1),
                ax=ax,
            )
        return ax

    def plot(
        self,
        concentrations=None,
        min_conc=None,
        max_conc=None,
        figsize=(10, 10),
        ax=None,
        annotate=False,
        alpha=None,
        **kwargs,
    ):
        fig, ax = self.plot_axes(figsize=figsize, ax=ax)
        self.plot_cells(
            ax=ax,
            concentrations=concentrations,
            min_conc=min_conc,
            max_conc=max_conc,
            alpha=alpha,
            **kwargs,
        )
        return fig, ax


class TriVeinModel(TriLeafModel, _VeinModel):
    def __init__(
        self, gridsize, vein_base_coordinates, cells=None, parameters=None
    ) -> None:
        TriLeafModel.__init__(
            self, gridsize, cells=cells, parameters=parameters
        )
        _VeinModel.__init__(self, vein_base_coordinates)

    def __reduce__(self):
        return (
            self.__class__,
            (
                self.gridsize,
                self.get_coordinates_of_index(self._vein_base),
                self.cells,
                self.parameters,
            ),
        )


class SquareVeinModel(SquareLeafModel, _VeinModel):
    def __init__(
        self, gridsize, vein_base_coordinates, cells=None, parameters=None
    ) -> None:
        SquareLeafModel.__init__(
            self, gridsize, cells=cells, parameters=parameters
        )
        _VeinModel.__init__(self, vein_base_coordinates)

    def __reduce__(self):
        return (
            self.__class__,
            (
                self.gridsize,
                self.get_coordinates_of_index(self._vein_base),
                self.cells,
                self.parameters,
            ),
        )


class HexVeinModel(HexLeafModel, _VeinModel):
    def __init__(
        self, gridsize, vein_base_coordinates, cells=None, parameters=None
    ) -> None:
        HexLeafModel.__init__(
            self, gridsize, cells=cells, parameters=parameters
        )
        _VeinModel.__init__(self, vein_base_coordinates)

    def __reduce__(self):
        return (
            self.__class__,
            (
                self.gridsize,
                self.get_coordinates_of_index(self._vein_base),
                self.cells,
                self.parameters,
            ),
        )


class TriStomataModel(TriLeafModel, _StomataModel):
    def __init__(
        self, gridsize, vein_base_coordinates, cells=None, parameters=None
    ) -> None:
        TriLeafModel.__init__(
            self, gridsize, cells=cells, parameters=parameters
        )
        _StomataModel.__init__(self, vein_base_coordinates)

    def __reduce__(self):
        return (
            self.__class__,
            (
                self.gridsize,
                self.get_coordinates_of_index(self._vein_base),
                self.cells,
                self.parameters,
            ),
        )


class SquareStomataModel(SquareLeafModel, _StomataModel):
    def __init__(
        self, gridsize, vein_base_coordinates, cells=None, parameters=None
    ) -> None:
        SquareLeafModel.__init__(
            self, gridsize, cells=cells, parameters=parameters
        )
        _StomataModel.__init__(self, vein_base_coordinates)

    def __reduce__(self):
        return (
            self.__class__,
            (
                self.gridsize,
                self.get_coordinates_of_index(self._vein_base),
                self.cells,
                self.parameters,
            ),
        )


class HexStomataModel(HexLeafModel, _StomataModel):
    def __init__(
        self, gridsize, vein_base_coordinates, cells=None, parameters=None
    ) -> None:
        HexLeafModel.__init__(
            self, gridsize=gridsize, cells=cells, parameters=parameters
        )
        _StomataModel.__init__(
            self, vein_base_coordinates=vein_base_coordinates
        )

    def __reduce__(self):
        return (
            self.__class__,
            (
                self.gridsize,
                self.get_coordinates_of_index(self._vein_base),
                self.cells,
                self.parameters,
            ),
        )
